//Ship class:
function Ship(shipjson) {
    this.sprite = new jaws.Sprite({
        image: shipjson.gfx, 
        x: 0, 
        y: 0, 
        angle:shipjson.location.angle,
        anchor: "center"
    });
    delete shipjson["gfx"];
    $.extend(this, shipjson);
    ///
    this.v = $V([0,0]);

    //Set up the officers
    //TODO: make this not nasty.
    var new_officers = {};
    for (var i = 0; i < this.officers.length; i++) {
        var new_officer = {};
        new_officer.ident = this.officers[i]; //"captain", "pilot", etc.
        new_officer.ship = this; //make ship accessible
        new_officer.tasks = [];
        new_officer.messages = [];

        new_officer.no_tasks = function(){} //do nothing by default
        new_officer.process_message = function(){} //do nothing by default
        new_officer.perform_task = function() {
            //Process tasks
            if (this.tasks[0] !== null && this.tasks[0] !== undefined) {
                var out = this.tasks[0].call(this.ship);
                if (out === 1) { //task completed
                    this.ship.clear_current_task(officer);
                }
            }
            else {
                this.no_tasks(); //Run a function if there aren't any tasks. Can be overwritten with officer-specific code..
            }

            //Process a single message
            if (this.messages[0] !== null && this.messages[0] !== undefined) {
                this.process_message();
            }
        }

        if (this.officers[i] === "captain") {
            new_officer.ship_has_goal = false; //default
            new_officer.no_tasks = function() {
                //TODO: figure out how to properly inject a new task when the global task (e.g. trade with planet) has been completed. Right now, this is via callbacks, but that's kinda iffy....
                if (this.ship_has_goal == false ) {
                    //Ask the Trader for a good trading route.
                    this.ship.postMessage("captain", "trading", "FIND_TRADE_ROUTE", undefined, undefined, function(body){
                        //Received an answer via callback
                        console.log("RECEIVED TRADE ROUTE: from '" + body.fromPlanet.name  + "' to '" + body.toPlanet.name + "'.");
                        this.ship.postMessage("captain", "pilot", "MOVE_TO_PLANET", body.fromPlanet, undefined, function(){
                            this.ship.postMessage("captain", "pilot", "MOVE_TO_PLANET", body.toPlanet, undefined, function(){
                                this.ship_has_goal = false; //We don't have a goal anymore...
                            });
                        });
                    });
                    this.ship_has_goal = true;
                }
            }
        }       

        if (this.officers[i] ===  "trading") {
            new_officer.process_message = function() {
                var msg = this.messages.shift();
                if (msg.messageType === "FIND_TRADE_ROUTE") {
                    console.log("TRADER: finding trade route");
                    //TODO: calculate trade route
                    var tradeRoute = {"fromPlanet": planets[parseInt(Math.random()*planets.length)], "toPlanet": planets[parseInt(Math.random()*planets.length)]};
                    //Respond to the requestor.
                    msg.callback.call(this.ship.officers[msg.fromOfficer], tradeRoute);
                }
            }
        }

        if (this.officers[i] === "pilot") {
            new_officer.process_message = function() {
                var msg = this.messages.shift();
                if (msg.messageType === "MOVE_TO_PLANET") {
                    //This message generates a few tasks for the Pilot:
                    //1) Move to target planet
                    //2) Land on the planet (fake animation)
                    //3) Wait a while, 'simulating' actions whilst on the planet (faked)
                    //4) Lift off again
                    //5) Tell whoever asked the pilot to go here that we're done..
                    var tgtplanet = msg.body;
                    //Push move to planet task
                    this.ship.Taskpush_Pilot_MoveTo(tgtplanet.location.position);
                    //Push fake hacky landing animation
                    this.tasks.push(function(){
                        this.sprite.alpha = Math.max(0,this.sprite.alpha - 0.01);
                        if (this.sprite.alpha <= 0) {
                            shipLanded(this, tgtplanet);
                            return 1;
                        } else {
                            return 0;
                        }
                    });
                    //Wait a bit
                    //TODO: should do some callback and tell all officers that we've landed so that they can do whatever they need to do...
                    var j = 100;
                    this.tasks.push(function(){
                        j = j - 1;
                        if (j <= 0) {
                            return 1;
                        } else {
                            return 0;
                        }
                    });
                    //Push fake liftoff animation
                    this.tasks.push(function(){
                        this.sprite.alpha = Math.min(this.sprite.alpha + 0.01,1);
                        if (this.sprite.alpha >= 1) {
                            return 1;
                        } else {
                            return 0;
                        }
                    });
                    //Done! Now do the callback to tell the captain we're done...
                    this.tasks.push(function(){
                        //this = ship!!!
                        msg.callback.call(this.officers[msg.fromOfficer], "done");
                        return 1;
                    });
                }
            }
        }
        
        new_officers[this.officers[i]] = new_officer;
    }
    this.officers = new_officers;
    
    ///
    this.location.angle = $V([Math.sin(this.location.angle),Math.cos(this.location.angle)*-1]); //Fix the angle; convert from angle in rads to Vector
    this.location.position = $V([this.location.Ux, this.location.Uy]);
    delete this.location.Ux;
    delete this.location.Uy;
    //////////////////////////////////////

    var NEXT_CORRELATIONID = 0;
    this.postMessage = function(fromOfficer, toOfficer, messageType, body, correlationID, callback) {
        if (correlationID == undefined) {
            correlationID = NEXT_CORRELATIONID;
            NEXT_CORRELATIONID += 1;
        }
        var msg = {
            "fromOfficer": fromOfficer,
            "toOfficer": toOfficer,
            "messageType": messageType,
            "body": body,
            "correlationID": correlationID,
            "callback": callback
        }
        if (this.officers[toOfficer] !== undefined) {
            this.officers[toOfficer].messages.push(msg);
        } else {
            //Ignore; message will die.
        }
    }

    //Loop through all officers and tell them to perform their task.    
    this.perform_tasks = function() {
        for (officer in this.officers) {
            this.officers[officer].perform_task();
        }
    }
    

    //////////////////// TASK FUNCTIONS //////////////////
    //TODO: these should be placed under their relevant officer!

    //Tell an officer to clear his current task
    this.clear_current_task = function(officer) {
        this.officers[officer].tasks.shift(); //pop from front
    }
    
    //Convenience method that tells the pilot to move to a position in the universe.
    this.Taskpush_Pilot_MoveTo = function(universe_vector) {
        this.officers["pilot"].tasks.push(this.Task_turnToMovement);
        this.officers["pilot"].tasks.push(this.Task_brake);
        this.officers["pilot"].tasks.push(function(){
            this.target = universe_vector;
            return 1
        });
        this.Taskpush_Pilot_TurnTo(universe_vector.subtract(this.location.position));
        this.officers["pilot"].tasks.push(function(){
            this.location.angle = universe_vector.subtract(this.location.position);
            return 1;
        }); //messy hack; should not be needed.
        this.officers["pilot"].tasks.push(this.Task_accelerate);
        this.officers["pilot"].tasks.push(this.Task_brake);
    }
    
    //Slow the ship down to 0...
    this.Taskpush_Pilot_Brake = function() {
        this.officers["pilot"].tasks.push(this.Task_turnToMovement);
        this.officers["pilot"].tasks.push(this.Task_brake);
    }

    //Turn the ship towards a vector
    this.Taskpush_Pilot_TurnTo = function(vector_angle) {
        this.officers["pilot"].tasks.push(function(){
            return this.turnTo(vector_angle)
        });
    }

    //Turn towards the current movement vector
    //Return 1 if succeeded
    this.Task_turnToMovement = function() {
        if (this.v.modulus() == 0) {
            return 1;
        }
        var b = this.turnTo(this.v);
        return b;
    }
    
    //Accelerate towards target (universe vector) until we need to break to prevent overshooting it 
    this.Task_accelerate = function(target) {
        if (target === undefined) {
            target = this.target;
        }
        var spd = Math.sqrt((this.v.elements[0] * this.v.elements[0]) + (this.v.elements[1] * this.v.elements[1]));
        var dst = (0-(spd*spd)) / (2*-this.movement.thrust);
        var cdst = this.location.position.subtract(target).modulus();
        //console.log("SPD:" + spd + "   DST:" + dst + "   CDST:" + cdst);
        if (cdst - dst < 50) {
            return 1;
        } else {
            this.thrust(1); 
            return 0;
        }
    }   

    //Decelerate until speed = 0 (won't work if ship isn't pointing at current movement vector)
    this.Task_brake = function() {
        var spd = Math.sqrt((this.v.elements[0] * this.v.elements[0]) + (this.v.elements[1] * this.v.elements[1]));
        return this.cruise(0,spd);
    }

    ///////////////// MISC FUNCTIONS ////////////////
    
    //Point at a coordinate in the universe
    this.pointat = function(x,y) {
        var tempv = $V([x,y]);
        this.pointatvu(tempv);
    }
    //Point at a vector (universe coordnate)
    this.pointatvu = function(vector) {
        this.location.angle = vector.subtract(this.location.position);
    }
    //Pount alongside a vector
    this.pointatv = function(vector) {
        this.location.angle = vector;
    }
    
    //Set angle to vector
    this.setvangle = function(x,y) {
        this.location.angle = $V([x,y]);
    }
    
    this.draw = function() {
        //Fix the sprite coordinates based on Universe coordinates and the player's position in the universe        
        this.sprite.x = this.location.position.getUx() - player.Ux + player.x;
        this.sprite.y = this.location.position.getUy() - player.Uy + player.y;      
        
        //Fix the angle based on the vector
        this.sprite.angle = (Math.atan2(this.location.angle.elements[1],this.location.angle.elements[0]) + DEG90_IN_RAD) * RAD_TO_DEG; //VERIFIED TO WORK
        
        //Draw the bastard if it's onscreen
        if (this.sprite.x > -100 && this.sprite.x <= jaws.width+100 && this.sprite.y > -100 && this.sprite.y <= jaws.height+100) {
            this.sprite.draw();
        }   
    }
    
    this.process = function() {  
        this.perform_tasks();
        this.location.position = this.location.position.add(this.v);
    }
    
    //Thrust; mult >= -1 && mult <= 1
    this.thrust = function(mult) {
        if (mult === undefined || mult === null) {
            mult = 1;
        }
        var thrustvec = this.location.angle.dup().toUnitVector().multiply(this.movement.thrust * mult);
        var nvec = this.v.add(thrustvec);
        var speed = nvec.elements[0]*nvec.elements[0]+nvec.elements[1]*nvec.elements[1];
        if (speed > this.movement.maxspeed) {
            nvec.elements[0] = (nvec.elements[0] / Math.sqrt(speed))*Math.sqrt(this.movement.maxspeed);
            nvec.elements[1] = (nvec.elements[1] / Math.sqrt(speed))*Math.sqrt(this.movement.maxspeed);
        }
        this.v = nvec;
    }
    
    //Turns left or right; dir >= -1 && dir <= 1
    this.turn = function(dir) {
        if (dir < -1) {
            dir = -1;
        }
        if (dir > 1) {
            dir = 1;
        }
        this.location.angle = this.location.angle.rotate(this.movement.turn_ratio*DEG_TO_RAD*dir, ZEROVEC);
    }
    
    //Turn towards a certain angle
    //Return 1 if turning not needed
    this.turnTo = function(vector_angle) {
        var ta = this.location.angle.angleTo(vector_angle);
        //console.log(vector_angle);
        //console.log(ta + "  ::  " + (this.movement.turn_ratio*DEG_TO_RAD));
        if (ta <= -this.movement.turn_ratio*DEG_TO_RAD) {
            //Turn left..
            this.turn(-1);
            return 0;
        } else if (ta >= this.movement.turn_ratio*DEG_TO_RAD ) {
            //Turn right..
            this.turn(1);
            return 0;
        } else {
            //Done!
            this.pointatv(vector_angle); //Guarantee fixation
            return 1;
        }
    }
    
    //Try to maintain a certain speed
    this.cruise = function(wantedspeed, currentspeed) {
        if (currentspeed - wantedspeed < 0.75 && currentspeed - wantedspeed > -0.75) {
            if (wantedspeed == 0) {
                this.v = $V([0,0]);
            } else {
            //do nothing
            }
            return 1;
        } else {
            if (wantedspeed < currentspeed) {
                this.thrust(-1);
            } else {
                this.thrust(1);
            }
            return 0;
        }
    }
}
