var $d = (function($d){

    $d.RESERVEDCHARS = ["!", "=", "<", ">", "/", "*", "-", "+", "~"];
    $d.DEBUGGING_ENABLED = false;
    $d._TEXTBUFFER = [];
    $d._CONVERSATIONSTATE = 0;

    $d.startConversation = function() {
        $d._CONVERSATIONSTATE = 1;
        for (var tp in $d.SCOPE.temporary) {  //Clear temporary scope
            $d.SCOPE.temporary[tp] = 0;
        }
        $d.executeBlock($d.getBlock("Greeting"));
    }

    $d.log = function (prefix, inLine) {
        if ($d.DEBUGGING_ENABLED) {
            console.log(">" + prefix + ">  " + inLine);
        }
    }

    $d.logFunc = function (inLine) {
        $d.log("LF", inLine);
    }

    $d.nullFunc = function (inLine) {
        $d.log("NULL", inLine);
    }

    $d._pushToObj = function (inLine) {
        $d._currentObj.lines.push(inLine);
    }

    $d.BLOCK_Greeting = function (inLine) {
        $d.log("GREETING", inLine);
        $d._pushToObj(inLine);
    }
    $d.BLOCK_GreetingStart = function (inLine) {
        $d.log("START_GREETING", inLine);
        $d.getNewObj("Greeting");
    }

    $d.BLOCK_Node = function (inLine) {
        $d.log("NODE", inLine);
        $d._pushToObj(inLine);
    }
    $d.BLOCK_NodeStart = function (inLine) {
        $d.log("START_NODE", inLine);
        $d.getNewObj("Node");
        $d._currentObj.options = [];
    }

    $d.BLOCK_Option = function (inLine) {
        $d.log("OPTION", inLine);
        $d._pushToObj(inLine);
    }
    $d.BLOCK_OptionStart = function (inLine) {
        $d.log("START_OPTION", inLine);
        $d.getNewObj("Option");
        $d.getLastNode().options.push($d._currentObj);    
    }

    $d.BLOCK_Text = function (inLine) {
        $d.log("TEXT", inLine);
        $d._pushToObj(inLine);
    }
    $d.BLOCK_TextStart = function (inLine) {
        $d.log("START_TEXT", inLine);
        $d.getNewObj("Text");
    }

    $d.endBlock = function (inLine) {
        $d._currentFunc = nullFunc;
    }

    $d.trimF = function (inLine) {
        return $.trim(inLine);
    }

    $d.getLastNode = function () {
        for (var i = $d._objects.length - 1; i >= 0; i--) {
            if ($d._objects[i].type == "Node") {
                return $d._objects[i];
            }
        }
        return null;
    }

    $d.setCurrentBlock = function (blockObj) {
        $("*.currentNode").removeClass("currentNode");
        $("*[data-uid=" + blockObj._UID + "]").addClass("currentNode");
    }

    $d._currentFunc = $d.nullFunc;
    $d._objects = [];
    $d._currentObj = null;
    $d._objID = 0;

    $d.SCOPE = {};
    $d.resetValues = function () {
        $d.SCOPE.global = {};
        $d.SCOPE.temporary = {};
        $d.SCOPE.namespaced = {};
    }
    $d.resetValues();
    
    $d.getNewObj = function (type) {
        $d._currentObj = {};    
    
        $d._objID += 1;
        $d._currentObj._UID = $d._objID;
        $d._currentObj.lines = [];
        $d._currentObj.type = type;
        if (type != "Option") {
            $d._objects.push($d._currentObj);
        }
        return $d._currentObj;
    }

    $d.firstPass = function (inArray) {
        for (i in inArray) {
            var line = inArray[i];
            line = $d.trimF(line); //Strip whitespace
            line = $.trim(line.substring(0,(line.indexOf("//") == -1?line.length : line.indexOf("//")))); //Strip comments + any potential whitespace (again)
        
            if (line.indexOf("greeting") == 0) {
                $d.BLOCK_GreetingStart(line);
                $d._currentFunc = $d.BLOCK_Greeting;
            }
            if (line.indexOf("option") == 0) {
                $d.BLOCK_OptionStart(line);
                $d._currentFunc = $d.BLOCK_Option;
            }
            if (line.indexOf("node") == 0) {
                $d.BLOCK_NodeStart(line);
                $d._currentFunc = $d.BLOCK_Node;
            }
            if (line.indexOf("text") == 0) {
                $d.BLOCK_TextStart(line);
                $d._currentFunc = $d.BLOCK_Text;
            }
        
            if (line.indexOf(".") == 0) {
                $d.execCurrentFunc(line);
                $d._currentFunc = $d.nullFunc;
            }
        
            $d.execCurrentFunc(line);
        }
    }

    /*
 * Does some checking, then executes the 'current' function....
 */
    $d.execCurrentFunc = function (line) {
        if (line != "") {
            $d._currentFunc(line);
        }
    }

    /**
 * Applies an array of consequences.
 */
    $d.applyConsequences = function (consequences) {
        if (consequences != undefined) {
            for (var i in consequences) {
                var cons = consequences[i];
                $d.evalConsequence(cons);
            }
        }
    }


    /**
 * Evaluates a string encased in 'square brackets', such as [foo] or [Credits+100]
 * 
 * According to the docs, this should do the following (in this order, presumably stopping once a valid action is found):
 * 1) find any arithmetic operations, then evaluate numeric parameters on either side and performing the arithmetic. NOTE: see above in numeric flags about arithmetic and order of operations.
 * 2) substitute text items with the named key. If the key is found but conditions are not met, that's treated as a match yielding the empty string, not a failure to match.
 * 3) substitute the value of any numeric flag with that key, or 0 if none.
 */
    $d.evalSquareBracketString = function (instring) {
        //instring still contains the square brackets
        instring = instring.substr(1,instring.length-2); // not anymore ;)
        var retstring = "";
    
        var spFound = false;
        for (var i = 0; i < instring.length; i++) {
            if ($d._isReservedChar(instring[i])) {
                spFound = true;
            }
        }
        var getBlockTest = $d.getBlock("Text", instring);
    
    
        //Case 1:
        if (spFound) {
            console.log(" TODO: evalSquareBracketString case 1");
        } else if (getBlockTest != null) {
            retstring = "";
            for (var j in getBlockTest.text) {
                retstring = retstring + getBlockTest.text[j];
                if (j < getBlockTest.text.length-1) {
                    retstring = retstring + "\n";
                }
            }
            //Now apply the consequences..
            if (getBlockTest.consequences != undefined) {
                $d.applyConsequences(getBlockTest.consequences);
            }
        } else {
            retstring = $d.getFlagValue($d.getScope(instring) + instring);
        }
    
        return retstring;
    }

    $d.displayText = function (textArr) {
        //TODO: evaluating TEXT NODE references.
        if (textArr != undefined) {
            for (var i in textArr) {
                var dispString = textArr[i];
                var sqBrackets = textArr[i].match(/\[[^\[\]]*\]/gi); //get all 'square bracket notation' strings
                for (j in sqBrackets) {
                    dispString = dispString.replace(sqBrackets[j], $d.evalSquareBracketString(sqBrackets[j]));
                }
                $.each(dispString.split("\n"), function(sa, sb){
                    console.log("> "  + sb);
                    $d._TEXTBUFFER.push(sb);
                });
            }
        }
    } 

    $d.gotoBlock = function (gotoID) {
        if (gotoID != undefined) {
            if (gotoID == "fight") {
                console.log(">> End of conversation (BATTLE INITIATED!)");
                $d._CURRENTOPTIONS = [];
                $d._CONVERSATIONSTATE = 3;
            } else if (gotoID == "done") {
                console.log(">> End of conversation (no battle)");
                $d._CURRENTOPTIONS = [];
                $d._CONVERSATIONSTATE = 2;
            } else {
                //Checking conditions is done in getBlock...
                $d.executeBlock($d.getBlock("Node", gotoID));
            }
        }
    }

    //From a set of possible options, check which ones are available to the player (eval their conditions)
    $d.getAvailableOptions = function (options) {
        var retArr = [];
        for (var o in options) {
            var option = options[o];
            if ($d.evalConditions(option.conditions)) {
                var addO = true;
                if (option.name != undefined) {
                    //Check if we already have a option in the array with this option's name
                    for (var i in retArr) {
                        if (retArr[i].name == option.name) {
                            addO = false;
                        }
                    }
                } 
                if (addO) { //add it (if no conflict)
                    retArr.push(option);
                }
            }
        }
        return retArr;
    }

    $d.dO = function (nr) {
        $d.applyConsequences($d._CURRENTOPTIONS[nr].consequences);
        $d.displayText($d._CURRENTOPTIONS[nr].text);
        $d.gotoBlock($d._CURRENTOPTIONS[nr]["goto"]);
    }

    $d._CURRENTOPTIONS = [];


    $d.executeBlock = function (block) {
        $d._currentObject = block;
        if ($d.DEBUGGING_ENABLED) {
            console.log("--Executing block--");
            console.log(block);
        }
        if (block.consequences != undefined) {
            $d.applyConsequences(block.consequences);
        }
        if (block.text != undefined) {
            $d.displayText(block.text);
        }
        if (block.options != undefined) {
            $d._CURRENTOPTIONS = $d.getAvailableOptions(block.options);
            for (var i in $d._CURRENTOPTIONS) {
                console.log(i + ": " + $d._CURRENTOPTIONS[i].player_text);
            }
        }
        console.log("-------------------");
        if (block["goto"] != undefined) {
            $d.gotoBlock(block["goto"]); //Should be the last line in this function!
        }
    }

    /*
 * Parse Greeting objects
 */
    $d.greetingPass = function () {
        for (o in $d._objects) {
            var object = $d._objects[o];
            if (object.type == "Greeting") {
                //Greeting objects have names, consequences AND conditions. All of those are placed in the FIRST line, together with the 'greeting' declaration.   
                $d._parseNCCs(object, object.lines[0].substring(9));
                object.lines = object.lines.slice(1);
                object["goto"] = object.lines.pop().substring(1);
                object.text = object.lines;
                delete object.lines;            
            }
        }
    }

    /*
 * Parse Text objects
 */
    $d.textPass = function () {
        for (o in $d._objects) {
            var object = $d._objects[o];
            if (object.type == "Text") {
                //Text objects have names, consequences AND conditions. All of those are placed in the FIRST line, together with the 'text' declaration.   
                $d._parseNCCs(object, object.lines[0].substring(5));
                object.lines = object.lines.slice(1);
                object["goto"] = object.lines.pop().substring(1);
                object.text = object.lines;
                delete object.lines;    
            }
        }
    }

    /*
 * Parse Option objects (which are stored inside Node objects!)
 */
    $d.optionPass = function () {
        for (o1 in $d._objects) {
            var pobject = $d._objects[o1];
            if (pobject.type == "Node") {
                for (o in pobject.options) {
                    var object = pobject.options[o];
                    //Option objects have names, consequences AND conditions. All of those are placed in the FIRST line, together with the 'option' declaration.                
                    $d._parseNCCs(object, object.lines[0].substring(7));
                    object.lines = object.lines.slice(1); //remove first line
                    object["player_text"] = object.lines[0];
                    object.lines = object.lines.slice(1); //remove first line
                    object["goto"] = object.lines.pop().substring(1);
                    object.text = object.lines;
                    delete object.lines;    
                }
            }
        }
    }

    /*
 * Parse Node objects
 */
    $d.nodePass = function() {
        for (o in $d._objects) {
            var object = $d._objects[o];
            if (object.type == "Node") {
                //Node objects have names, consequences AND conditions. All of those are placed in the FIRST line, together with the 'node' declaration.    
                $d._parseNCCs(object, object.lines[0].substring(5));
                object.lines = object.lines.slice(1); //remove first line
                //That should be it for Node objects..
                delete object.lines;
            }
        }
    }

    $d._isReservedChar = function (inchar) {
        return $.inArray(inchar,$d.RESERVEDCHARS) > -1;
    }

    $d._getScopedFlags = function (instring) {
        var flags = $d._getFlags(instring);
        var scopedFlags = $.merge([],flags);
        for (var i in scopedFlags) {
            var flag = scopedFlags[i];
            scopedFlags[i] = $d.getScope(flag) + flag;
        }
        return {
            "flags" : flags, 
            "scopedFlags": scopedFlags
        };
    }

    $d.evalCondition = function (unparsedExpression) {
        return eval($d._getParsedConditionString(unparsedExpression));
    }

    $d.evalConsequence = function (unparsedExpression) {
        return eval($d._getParsedConsequenceString(unparsedExpression));
    }

    $d._getParsedConsequenceString = function (instring) { //For eval
        var arr = $d._getParsedConsequence(instring);
        var retStr = "";
        if (arr.length > 1) {
            if (arr[0].value == "~") {
                retStr = arr[1].value + " = ($d.getFlagValue(\"" + arr[1].value + "\") == 0 ? 1 : 0)";
                if (arr.length > 2) {
                    console.warn("Warning: some arguments ignored for expression '" + instring + "' --> ~ expressions may only contain a tilde and a flag");
                }
            } else if (arr[0].value == "+") {
                if (arr.length > 3) {
                    retStr += arr[1].value + " ";
                    for (var i = 2; i < arr.length; i++) {
                        if (arr[i].type == "FLAG") {
                            retStr += "$d.getFlagValue(\"" + arr[i].value + "\") ";
                        } else {
                            retStr += arr[i].value + " ";
                        }
                    }
                } else {
                    retStr = arr[1].value + " = 1";
                }
            }
            else if (arr[0].value == "-") {
                retStr = arr[1].value + " = 0";
                if (arr.length > 2) {
                    console.warn("Warning: some arguments ignored for expression '" + instring + "' --> - expressions may only contain a minus and a flag");
                }
            }
        } else {
            console.warn("Ignored a consequence with less than 2 items ('" + instring + "') --> This shouldn't happen (or have any meaning whatsoever)..");
        }
        return retStr;
    }

    $d._getParsedConditionString = function (instring) { //For eval
        var arr = $d._getParsedCondition(instring);
        var retStr = "";
        if (arr.length > 0) {
            if (arr[0].value == "!") {
                retStr += arr.shift().value; //Pop the first element off
            } else if (arr[0].value == "==") {
                arr.shift(); //Ignore the "==" completely
            }
            retStr += "(";
            for (var i in arr) {
                if (arr[i].type == "FLAG") {
                    retStr += "$d.getFlagValue(\"" + arr[i].value + "\")";
                } else {
                    retStr += arr[i].value + " ";
                }
            }
            retStr += ")";
        }
        return retStr;
    }

    $d.getFlagValue = function (flagStrPtr) {
        var x = flagStrPtr;
        x = x.replace("$d.SCOPE.","");
        var sc = $d.SCOPE;
        var split = x.split(".")
        for (var i in split) {
            if (sc[split[i]] != undefined) {
                sc = sc[split[i]];
            } else {
                sc[split[i]] = 0;
                sc = sc[split[i]];
            }
        }
        return sc;
    }

    //Not very DRY, but fuck it.
    $d._getParsedConsequence = function (instring) {
        var arr = $d.__getParsedExpressionFirstPass(instring);
        for (var i in arr) {
            var a = arr[i];
            if (a.type == "FLAG") {
                a.value = $d.getScope(a.value) + "" + a.value; // Add the 'scope' stuff
            }
        }
        return arr;
    }

    //Not very DRY (see above), but fuck it.
    $d._getParsedCondition = function (instring) {
        var arr = $d.__getParsedExpressionFirstPass(instring);
        for (var i in arr) {
            var a = arr[i];
            if (a.type == "FLAG") {
                a.value = $d.getScope(a.value) + "" + a.value; // Add the 'scope' stuff
            }
            if (a.type == "RESERVED") {
                //For JavaScript eval: replace "=" with "==".
                a.value = a.value.replace("=", "==");
            }
        }
        return arr;
    }

    $d.__getParsedExpressionFirstPass = function (instring) {
        var retArr = [];
        var cObj = {};
        //don't push it
        for (var i in instring) {
            var ic = instring[i];
            if ($d._isReservedChar(ic)) {
                if (cObj.type === "RESERVED") {
                    cObj.value = cObj.value + ic;
                } else {
                    cObj = {};
                    retArr.push(cObj);
                    cObj.type = "RESERVED";
                    cObj.value = "" + ic;
                }
            } else if (ic >= "0" && ic <= "9") {
                if (cObj.type === "NUMBER") {
                    cObj.value = cObj.value + "" + ic;
                } else {
                    cObj = {};
                    retArr.push(cObj);
                    cObj.type = "NUMBER";
                    cObj.value = "" + ic;
                }
            } else {
                if (cObj.type === "FLAG") {
                    cObj.value = cObj.value + "" + ic;
                } else {
                    cObj = {};
                    retArr.push(cObj);
                    cObj.type = "FLAG";
                    cObj.value = "" + ic;
                }
            }
        }
        return retArr;
    }

    $d._getFlags = function (instring) {
        var retArr = [];
        retArr.push("");
        for (var i in instring) {
            var ic = instring[i];
            if ($d._isReservedChar(ic)) {
                if (retArr[retArr.length-1] != "") {
                    retArr.push("");
                }
            } else {
                retArr[retArr.length-1] =  retArr[retArr.length-1] + ic;
            }
        }
        for (var i in retArr) {
            if (retArr[i] != "") {
                if (retArr[i][0] >= "0" && retArr[i][0] <= "9") {
                    retArr.splice(i,1); //remove numbers or 'flags'  starting with a number (which aren't really flags)
                }
            }
        }
        if (retArr[retArr.length-1] == "") {
            retArr.pop(); //remove empty strings...
        }
        return retArr;
    }

    $d.evalConditions = function (conditions) {
        var proc = true;
        for (var i in conditions) {
            var cond = conditions[i];
            var evalled = $d.evalCondition(cond);
            proc = proc & evalled;
        }
        return proc;
    }

    $d.getBlock = function (blocktype, blockname) {
        for (var oi in $d._objects) {
            if ($d._objects[oi].type == blocktype && (blockname == undefined || $d._objects[oi].name == blockname)) {
                var b = $d._objects[oi];
                if ($d.evalConditions(b.conditions) == true) {
                    return b;
                }
            }
        }
        return null; //couldn't find a greeting
    }

    /**
*  Parses Name, Conditions and Consequences. Adds them to the object.
*  param object: the Object to store Conditions and Consequences on.
*  param inString: the String to parse (should not start with "greeting", "option" or "text")
*/
    $d._parseNCCs = function (object, inString) {
        object["consequences"] = [];
        object["conditions"] = [];
        args = inString.split(" ");
        for (ai in args) {
            var arg = $.trim(args[ai]);
            if (arg != "") {
                if ($d._isConsequence(arg)) {
                    object["consequences"].push(arg);
                } else if ($d._isCondition(arg)) {
                    object["conditions"].push(arg);
                } else if ($d._isCompound(arg)) {
                    var targ = arg.substring(1);
                    if (arg[0] == ">") {
                        object["conditions"].push("!" + targ);
                        object["consequences"].push("+" + targ);
                    } else if (arg[0] == "<") {
                        object["conditions"].push("=" + targ);
                        object["consequences"].push("-" + targ);
                    }
                } else { //must be a name..
                    if (object["name"] === undefined) {
                        object["name"] = arg;
                    } else {
                        console.warn("WARNING: Multiple 'name' arguments applied for block object! Erroring line is: '" + object.lines[0] + "'.. Additional argument (" + arg + ") ignored!");
                    }
                }
            }
        }
    }

    /**
* Returns true if the inString is a condition; false otherwise
*/
    $d._isCondition = function (inString) {
        switch(inString[0]) {
            case "=":
                return true;
            case "!":
                return true;
            default:
                return false;
        }
        return false;
    }

    /**
* Returns true if the inString is a consequence; false otherwise
*/
    $d._isConsequence = function (inString) {
        switch(inString[0]) {
            case "+":
                return true;
            case "-":
                return true;
            case "~":
                return true;
            default:
                return false;
        }
        return false;
    }

    /**
* Returns true if the inString is a compound operation; false otherwise
*/
    $d._isCompound = function (inString) {
        switch(inString[0]) {
            case ">":
                return true;
            case "<":
                return true;
            default:
                return false;
        }
        return false;
    }

    $d.uqmParse = function (input) {
        $d._objects = [];
        $d.resetValues();
    
        input = input.replace(/&gt;/g,">").replace(/&lt;/g,"<");
        input = input.split("\n");
    
        $d.firstPass(input);
        //Input isn't needed anymore!
        $d.greetingPass();
        $d.textPass();
        $d.optionPass();
        $d.nodePass();
    
        console.log("Parsing complete. The following objects have been read:");
        console.log($d._objects);
        console.log(" ----------------- ");
        console.log("To start the conversation, call startConversation() (and optionally resetValues() which also resets namespaced and global scopes)");
        console.log("To select an option, call dO(option_nr)");
        console.log("If you want additional debugging info, set $d.DEBUGGING_ENABLED to TRUE");
        console.log("If you want to inspect the state of the conversation, try looking at $d._objects, $d._currentObject, _CURRENTOPTIONS and SCOPE (which contains all stored values!)")
        console.log(" ----------------- ");
    }

    $d.getScope = function (input) {
        if (input.toLowerCase() == input) { 
            return "$d.SCOPE.temporary.";
        } else if (input.toUpperCase() == input) {
            return "$d.SCOPE.global."
        } else {
            return "$d.SCOPE.namespaced.";
        }
    }



    console.log("Initialized JS-Dialog parser...");
    console.log("Version 0.2");
    console.log(" Probably contains bugs and may not be 100% according to spec");
    console.log("Supports: ");
    console.log(" +Parsing conversation input completely");
    console.log(" +Rudimentary evaluation support (order of operations may not be according to spec)");
    console.log(" +Playing conversations in the JavaScript console");
    console.log("Doesn't support:");
    console.log(" -Proper order of operations");
    console.log(" -Arithmetics in block text items (like [Foo+bar-baz])");
    console.log("##################################################################################");
    console.log(" ");


    return $d;
})($d || {});