var radar;
var radar_enemies;
var RADAR_ZOOM = 0.05;
var RADAR_WIDTH = 200;
var RADAR_HEIGHT = 200;
var player;
var enemies;
var planets;
var DEG_TO_RAD = 0.0174532925;
var DEG90_IN_RAD = 1.57079633;
var RAD_TO_DEG = 57.2957795;
var parallax;
var NVEC = $V([0,1]);
var ZEROVEC = $V([0,0]);

//Override the getScope function
$d.getScope = function() {
    return "$d.SCOPE.";
}

Vector.prototype.getUx = function() {
    return this.elements[0];
}
Vector.prototype.getUy = function() {
    return this.elements[1];
}
Vector.prototype.angleTo = function(vector) {
    var sign = (this.elements[0] * vector.elements[1]) - (this.elements[1] * vector.elements[0]); //Based on cross product implementation and http://stackoverflow.com/questions/2663570/how-to-calculate-both-positive-and-negative-angle-between-two-lines
    var anglefrom = Math.abs(this.angleFrom(vector));
    if (anglefrom > Math.PI) {
        anglefrom = anglefrom - Math.PI;
    }
    if (sign < 0) {
        sign = -1;
    }
    if (sign >=  0) {
        sign = 1;
    }
    //console.log("sign: " + sign + "  anglefrom: " + anglefrom);
    return sign * anglefrom;
}
  
function testAbout(invar, expected) {
    return (invar > expected - 0.25 && invar < expected + 0.25);
}

function dTest(vec1, vec2, expected) {
    console.log(testAbout(vec1.angleTo(vec2)* RAD_TO_DEG,expected));
    console.log(testAbout(vec2.angleTo(vec1)* RAD_TO_DEG,-expected));
}
  
function test() {
    dTest(
        $V([1,0]),
        $V([0,1]),
        90
        );
    
    dTest(
        $V([0,1]),
        $V([-1,-1]),
        135
        );
             
    dTest(
        $V([-1,-1]),
        $V([0,-1]),
        45
        );
             
    dTest(
        $V([1,-1]),
        $V([1,1]),
        90
        );
             
             
             
    var v1 = $V([0,1]);
    var v2 = $V([0,-1]);
    console.log("perp: " + (testAbout(v1.angleTo(v2)* RAD_TO_DEG,180) ||testAbout(v1.angleTo(v2)* RAD_TO_DEG,-180)));
    console.log("----------------");
}

test();

//Planet class:
function Planet(planetjson) {
    this.sprite = new jaws.Sprite({
        image: planetjson.gfx, 
        x: 0, 
        y: 0, 
        angle:0,
        anchor: "center"
    }); 
    delete planetjson["gfx"];
    $.extend(this, planetjson);
    this.location.position = $V([this.location.Ux, this.location.Uy]);
    delete this.location.Ux;
    delete this.location.Uy;
    
    //Functions:
    this.draw = function() {
        //Fix the sprite coordinates based on Universe coordinates and the player's position in the universe        
        this.sprite.x = this.location.position.getUx() - player.Ux + player.x;
        this.sprite.y = this.location.position.getUy() - player.Uy + player.y;      

        //Draw the bastard if it's onscreen
        if (this.sprite.x > -1000 && this.sprite.x <= jaws.width+1000 && this.sprite.y > -1000 && this.sprite.y <= jaws.height+1000) {
            this.sprite.draw();
        }
    }
}


function forEnemy(funct) {
    for (var e in enemies) {
        funct(enemies[e]);
    }
}


function MyGameState() {
    
    
    function applyThrust() {
        var vec = $V([Math.cos((player.angle*DEG_TO_RAD)-DEG90_IN_RAD) * 0.1,Math.sin((player.angle*DEG_TO_RAD)-DEG90_IN_RAD) * 0.1]);
        var nvec = player.v.add(vec);
        var speed = nvec.elements[0]*nvec.elements[0]+nvec.elements[1]*nvec.elements[1];
        if (speed > player.cap) {
            nvec.elements[0] = (nvec.elements[0] / Math.sqrt(speed))*Math.sqrt(player.cap);
            nvec.elements[1] = (nvec.elements[1] / Math.sqrt(speed))*Math.sqrt(player.cap);
        }
        player.v = nvec;
    }
    
    function movePlayer() {
        player.Ux = player.Ux + player.v.elements[0];
        player.Uy = player.Uy + player.v.elements[1];
    }
    
    function startConversation(conversationID, convdata) {
        $.get("conversations/" + conversationID + ".txt", function(data){
            $d.uqmParse(data);
            $d.resetValues();
            if (data !== undefined) {
                $d.SCOPE = convdata;
            }
            $d.startConversation();
            displayConversationState();
        });
    }
    
    function displayConversationState() {
        var buttons = [];
        for (cOpt in $d._CURRENTOPTIONS) {
            var bopt = parseInt(cOpt);
            console.log("BOPT::: " + bopt);
            buttons.push({
                numid: 0,
                text:$d._CURRENTOPTIONS[cOpt].player_text, 
                type:"btn btn-info", 
                click:function($noty){
                    $noty.close();
                    $d.dO($(this).data("btnid"));
                    displayConversationState();
                }
            });
        }         
        
        var tText = $d._TEXTBUFFER.join("<br/>");
        
        var genT = $("#" + noty({
            type: "information", 
            layout: "bottomRight", 
            text: tText, 
            buttons:buttons,
            timeout: false,
            closable: true,
            closeButton: true
        }));
        
        if ($d._CONVERSATIONSTATE == 2) {
            setTimeout(function(){
                noty({
                    type: "warning", 
                    layout: "bottomRight", 
                    text: "Connection terminated", 
                    buttons:buttons,
                    timeout: 3000,
                    closable: true,
                    closeButton: true
                })
            },1000);
        }
        
        genT.find(".noty_buttons").css("text-align", "left").find("button").after("<br/>");
        var bTns = genT.find(".noty_buttons button");
        $.each(bTns, function(a,b){
            $(b).data("btnid",a);
        });
        $d._TEXTBUFFER = [];
    }
  
    
    
    /* Called once. Put your one-time initializing here. */
    this.setup = function() { 
        parallax = new jaws.Parallax({
            repeat_x: true, 
            repeat_y: true
        })
        parallax.addLayer({
            image: "parallax_1.png", 
            damping: 50
        })
        parallax.addLayer({
            image: "parallax_2.png", 
            damping: 30
        })
        parallax.addLayer({
            image: "parallax_3.png", 
            damping: 10
        })
        parallax.addLayer({
            image: "parallax_4.png", 
            damping: 1
        })
        
        player = new jaws.Sprite({
            image: "ship.png", 
            x: 0,
            y: 0,
            angle: 135,
            anchor:"center"
        });
        player.v = $V([0,0]);
        player.Ux = 0;
        player.Uy = 0;
        player.cap = 25;
        
        planets = [];
        
        enemies = [];
        $.get("planets.json", function(pdata){
            for (is in pdata) {
                planets.push(new Planet(pdata[is]));
            }
            $.get("ships.json", function(data){
                for (is in data) {
                    enemies.push(new Ship(data[is]));
                }
                //Generate more ships
                for (var i = 0; i < 100; i++) {
                    var shobj = {
                        "type": "ship",
                        "gfx": "ship" + Math.floor(Math.random()*14) + ".png",
                        "movement": {
                            "turn_ratio": 0.5 + (Math.random()*1.0),
                            "thrust": 0.01 + (Math.random()*0.10),
                            "maxspeed": 8 + (Math.random()*32)
                        },
                        "name": "Random Ship " + i,
                        "location": {
                            "Ux": -500 + Math.random() * 1000,
                            "Uy": -500 + Math.random() * 1000,
                            "angle": Math.random() * Math.PI
                        },
                        "conversation": null,
                        "officers": [
                        "captain", "trading", "pilot"
                        ]
                    };
                    enemies.push(new Ship(shobj));
                }
            });
        });
        


        $("#canvas").click(function(evt){
            var rect = jaws.Rect(evt.offsetX,evt.offsetY,1,1);
            rect.rect = function(){
                return this;
            } //fail Jaws
            for (i in planets) {
                if (jaws.collideOneWithOne(rect, planets[i].sprite)) {
                    var planet = planets[i];
                    new function() {
                        noty({
                            text: "Do you want to engage conversation with <b>" + planet.name + "</b>?", 
                            layout:"bottomRight",
                            type:"information",
                            buttons: [
                            {
                                type: 'button green btn btn-success', 
                                text: 'Ok', 
                                click: function($noty) {
                                    $noty.close();
                                    if (planet.conversation !== null) {
                                        startConversation(planet.conversation, planet);
                                    } else {
                                        noty({
                                            text: "The planet does not respond to your call for communication.", 
                                            layout: "bottomRight", 
                                            type: "warning",  
                                            closable: true, 
                                            closeButton: true, 
                                            timeout: 3000
                                        });    
                                    }
                                }
                            },
                            {
                                type: 'button pink btn btn-info', 
                                text: 'Cancel', 
                                click: function($noty) {
                                    $noty.close();
                                }
                            }
                            ],
                            closable: false,
                            timeout: false
                        });
                    }();
                }
            }            
            for (i in enemies) {
                if (jaws.collideOneWithOne(rect, enemies[i].sprite)) {
                    var enemy = enemies[i];
                    new function() {
                        noty({
                            text: "Do you want to engage conversation with <b>" + enemy.name + "</b>?", 
                            layout:"bottomRight",
                            type:"information",
                            buttons: [
                            {
                                type: 'button green btn btn-success', 
                                text: 'Ok', 
                                click: function($noty) {
                                    $noty.close();
                                    if (enemy.conversation !== null) {
                                        startConversation(enemy.conversation, enemy);
                                    } else {
                                        noty({
                                            text: "The ship does not respond to your call for communication.", 
                                            layout: "bottomRight", 
                                            type: "warning",  
                                            closable: true, 
                                            closeButton: true, 
                                            timeout: 3000
                                        });    
                                    }
                                }
                            },
                            {
                                type: 'button pink btn btn-info', 
                                text: 'Cancel', 
                                click: function($noty) {
                                    $noty.close();
                                }
                            }
                            ],
                            closable: false,
                            timeout: false
                        });
                    }();
                }
            }
        });
    }
  
    /* Called each gametick. Put your gamelogic here. */
    this.update = function() { 
        if(jaws.pressed("left"))  {
            player.rotate(-2);
            player.angle = player.angle % 360;
        }
        if(jaws.pressed("right")) {
            player.rotate(2);
            player.angle = player.angle % 360;
        }
        if(jaws.pressed("up"))    {
            applyThrust()
        }
        
        movePlayer();
        
        for (i in enemies) {
            enemies[i].process();
        //enemies[i].move();
        }
        
        
        //Parallax fixer
        parallax.camera_x = player.Ux+2147483648;
        parallax.camera_y = player.Uy+2147483648;
        
        //Player fixer
        player.x = jaws.width*0.5;
        player.y = jaws.height*0.5;
    }

    /* Called each gametick after update(). Put your drawing here. */
    this.draw = function() {        
        //Background:
        jaws.context.fillRect(0,0,jaws.width,jaws.height);
        parallax.draw();

        //Objects:
        for (i in planets) {
            planets[i].draw();
            if (planets[i].svg_obj === undefined) {
                planets[i].svg_obj = radar.circle(
                    0,
                    0,
                    5).attr({
                    "stroke": "black", 
                    "fill": "blue"
                });
                console.log("CREATE");
            }
            planets[i].svg_obj.attr({"cx": (planets[i].location.position.getUx()-player.Ux)*RADAR_ZOOM+RADAR_WIDTH, "cy": (planets[i].location.position.getUy()-player.Uy)*RADAR_ZOOM+RADAR_HEIGHT});
        }
        for (i in enemies) {
            if (enemies[i].svg_obj === undefined) {
                enemies[i].svg_obj = radar.circle(
                    0,
                    0,
                    1).attr({
                    "stroke": "red", 
                    "fill": "red",
                    "z-index": 123123
                });
                console.log("CREATE");
            }
            enemies[i].svg_obj.attr({"cx": (enemies[i].location.position.getUx()-player.Ux)*RADAR_ZOOM+RADAR_WIDTH, "cy": (enemies[i].location.position.getUy()-player.Uy)*RADAR_ZOOM+RADAR_HEIGHT});
            
            enemies[i].draw();
        }        
        
        if (player.svg_obj === undefined) {
            player.svg_obj = radar.circle(200,200,2).attr({
                "stroke": "black", 
                "fill": "#0F0"
            });
            console.log("CREATE");
        }
        
        player.draw(); //Draw last so it's on top
        
        $("#coords").html(Math.round(player.Ux) + " " + Math.round(player.Uy));
    }
}